library getworld;

import 'dart:convert';
import 'dart:developer';
import 'dart:io';
import './scr/xcountry.dart';

import './getworld.dart';

export './scr/xcity.dart';
export './scr/xcurrency.dart';
export './scr/xdialling.dart';
export './scr/xgeographical.dart';
export './scr/xlatlng.dart';
export './scr/xname.dart';
export './scr/xvat_rate.dart';
export './scr/xenums.dart';
export './scr/xlanguage.dart';
export './scr/xextra.dart';
export './scr/xpopulation.dart';
export './scr/xstate.dart';

class GetWorld {
  static bool _loaded = false;

  // bool countries = true;
  // bool currencies = true;
  // bool languages = true;

  GetWorld();

  factory GetWorld.intialize() {
    return GetWorld();
  }
  Future<void> initialize({
    bool countires = true,
    bool currencies = true,
    bool languages = true,
    bool name = true,
    bool alt_spellings = true,
    bool dialling = true,
    bool population = true,
    bool extra = true,
    bool geographical = true,
    bool states = true,
    bool cities = true,
  }) async {
           assert(!(states == false && cities == true), "getting cities need states to be true");
    if (_loaded == false) {
      if (languages == true) await _initial_Languages();
      if (currencies == true) await _initial_Currency();
      if (countires == true) await _initial_Countries();
      log("GetWorld initialized", name: "GetWorld");
      _loaded = true;
    } else {
      assert(_loaded != true, "GetWorld is aready intialized");
    }
  } 

  static final List<XCountry> _countries = [];
  List<XCountry> get Countries {
    return _countries;
  }

  static final List<XCurrency> _currencies = [];
  List<XCurrency> get Currencies {
    return _currencies;
  }

  static final List<XLanguage> _languages = [];
  List<XLanguage> get Languages {
    return _languages;
  }

  Future<void> _initial_Languages() async {
    List<dynamic> data = jsonDecode(await File("./jsons/languages.json").readAsString());
    // wirtting main languange informations
    _languages.addAll(data.map((e) => XLanguage(
          iso_639_1__alpha2: e["iso_639_1__alpha2"],
          iso_639_2__alpha3: e["iso_639_2__alpha3"],
          family: e["family"],
          name: e["name"],
          native: e["nativeName"],
          wikiUrl: e["wikiUrl"],
        )));

    // writting translations
    for (XLanguage language in _languages) {
      for (dynamic c in data) {
        if (c["iso_639_2__alpha3"] == language.iso_639_2__alpha3 && c["name_in"].length > 0) {
          Map<XLanguage, List<String>?> eng = {
            _languages.firstWhere((element) => element.iso_639_2__alpha3 == "ENG"):
                List<String>.from(c["name_in"]["ENG"])
          };
          Map<XLanguage, List<String>?> nld = {
            _languages.firstWhere((element) => element.iso_639_2__alpha3 == "NLD"):
                List<String>.from(c["name_in"]["NLD"])
          };
          Map<XLanguage, List<String>?> fra = {
            _languages.firstWhere((element) => element.iso_639_2__alpha3 == "FRA"):
                List<String>.from(c["name_in"]["FRA"])
          };

          language.name_in!.addAll(eng);
          language.name_in!.addAll(nld);
          language.name_in!.addAll(fra);
        }
      }
    }

    log("Languages initialized", name: "GetWorld");
  }

  Future<void> _initial_Currency() async {
    List<dynamic> data = jsonDecode(await File("./jsons/currencies.json").readAsString());
    _currencies.addAll(List<XCurrency>.from(data.map((currency) => XCurrency(
          iso_4217_code: currency["iso_4217_code"],
          iso_4217_numeric: currency["iso_4217_numeric"].toString(),
          iso_4217_name: currency["iso_4217_name"],
          iso_4217_minor_unit: currency["iso_4217_minor_unit"],
          withdrawal_date: currency["withdrawal_date"],
          decimal_digits: currency["decimal_digits"],
          full_name: currency["full_name"],
          name_plural: currency["name_plural"],
          rounding: currency["rounding"],
          symbol: currency["symbol"],
          symbol_native: currency["symbol_native"],
        ))));

    log("Currencies initialized", name: "GetWorld");
  }

  Future<void> _initial_Countries([
    bool alt_spellings = true,
    bool dialling = true,
    bool population = true,
    bool extra = true,
    bool geographical = true,
    bool states = true,
    bool cities = true,
  ]) async {
    List<dynamic> data = jsonDecode(await File("./jsons/data_compact.json").readAsString());

    _countries.addAll(data
        .map((country) => XCountry(
            iso_3166_1_alpha2: country["iso_3166_1_alpha2"],
            iso_3166_1_alpha3: country["iso_3166_1_alpha3"],
            iso_3166_1_numeric: country["iso_3166_1_numeric"],
            cioc: country["cioc"],
            name: _Name(country),
            natives: _Natives(country),
            alt_spellings: (alt_spellings == false)? null : List<String>.from(country["alt_spellings"]),
            translations: (_languages.length == 0) ? null : _translations(country),
            capital: List<String>.from(country["capital"]),
            unMember: country["unMember"],
            status: country["status"],
            flag_symbol: country["flag_symbol"],
            tld: List<String>.from(country["tld"]),
            dialling:(dialling == false)? null : _Dialling(country),
            population:(population == false)? null : _Population(country),
            extra: (extra == false)? null : _Extra(country),
            geographical: (geographical == false)? null : _Geographical(country),
            languages: (_languages.length == 0) ? null : _Languages(country),
            currencies: (_currencies.length == 0) ? null : _Currencies(country),
            states: (states == false)? null : _States_Cities(country, cities)))
        .toList());

    //Fill borders
    for (var country in _countries) {
      country.geographical!.borders = data
          .singleWhere((element) => element["iso_3166_1_alpha3"] == country.iso_3166_1_alpha3)["geographical"]
              ["borders"]
          .map<XCountry>((b) => _countries.singleWhere((c) => c.iso_3166_1_alpha3 == b))
          .toList();
    }

    log("Countries initialized", name: "GetWorld");
  }

  List<XCurrency>? _Currencies(country) {
    return country["currencies"]
        .map<XCurrency>((cur) => _currencies.singleWhere((element) => element.iso_4217_code == cur))
        .toList();
  }

  List<XLanguage>? _Languages(country) {
    return country["languages"]
        .map<XLanguage>((lan) => _languages.singleWhere((element) => element.iso_639_2__alpha3 == lan))
        .toList();
  }

  List<XState>? _States_Cities(country,[bool cities = true]) {
    return country["states"]
        ?.map<XState>((s) => XState(
              id: s["id"],
              name: s["name"],
              state_code: s["state_code"],
              type: s["type"],
              latLng: XLatLng(latitude: s["latitude"], longitude: s["longitude"]),
              cities: (cities == false) ? null :s["cities"]
                  ?.map<XCity>((c) => XCity(
                        id: c["id"],
                        name: c["name"],
                        latLng: XLatLng(latitude: c["latitude"], longitude: c["longitude"]),
                      ))
                  .toList(),
            ))
        .toList();
  }

  Map<XLanguage, XName>? _translations(dynamic country) {
    if (country == null || _languages.isEmpty) return null;
    return Map<XLanguage, XName>.from(country["translations"].map<XLanguage, XName>((key, value) =>
        MapEntry<XLanguage, XName>(_languages.firstWhere((element) => element.iso_639_2__alpha3 == key.toUpperCase()),
            XName(official: value["official"], common: value["common"]))));
  }

  Map<XLanguage, XName>? _Natives(dynamic country) {
    if (country == null || _languages.isEmpty) return null;
    return Map<XLanguage, XName>.from(country["natives"].map<XLanguage, XName>((key, value) =>
        MapEntry<XLanguage, XName>(_languages.firstWhere((element) => element.iso_639_2__alpha3 == key.toUpperCase()),
            XName(official: value["official"], common: value["common"]))));
  }

  XDialling _Dialling(country) {
    return XDialling(
        calling_code: country["dialling"]["calling_code"],
        national_number_lengths: country["dialling"]["national_number_lengths"]);
  }

  XPopulation _Population(country) {
    return XPopulation(
        count: country["population"]["count"], worldPercentage: country["population"]["worldPercentage"]);
  }

  XName _Name(dynamic country) {
    return XName(
      official: country["name"]["official"],
      common: country["name"]["common"],
    );
  }

  XExtra _Extra(dynamic country) {
    return XExtra(
      geonameid: country["extra"]["geonameid"],
      edgar: country["extra"]["edgar"].toString(),
      itu: country["extra"]["itu"].toString(),
      marc: country["extra"]["marc"].toString(),
      wmo: country["extra"]["wmo"].toString(),
      ds: country["extra"]["ds"].toString(),
      fifa: country["extra"]["fifa"].toString(),
      fips: country["extra"]["fips"].toString(),
      gaul: country["extra"]["gaul"],
      ioc: country["extra"]["ioc"].toString(),
      cowc: country["extra"]["cowc"].toString(),
      cown: country["extra"]["cown"],
      fao: country["extra"]["fao"],
      imf: country["extra"]["imf"],
      ar5: country["extra"]["ar5"].toString(),
      address_format: country["extra"]["address_format"],
      eu_member: country["extra"]["eu_member"].toString(),
      // vat_rates: c["extra"]["vat_rates"]      <- need to filed
    );
  }

  XGeographical _Geographical(dynamic country) {
    return XGeographical(
      latLng_dmc: XLatLng(
          latitude: country["geographical"]["latLng_dmc"]["latitude"],
          longitude: country["geographical"]["latLng_dmc"]["longitude"]),
      latLng_dec: XLatLng(
          latitude: country["geographical"]["latLng_dec"]["latitude"],
          longitude: country["geographical"]["latLng_dec"]["longitude"]),
      latLng_max: XLatLng(
          latitude: country["geographical"]["latLng_min"]["latitude"],
          longitude: country["geographical"]["latLng_min"]["longitude"]),
      latLng_min: XLatLng(
          latitude: country["geographical"]["latLng_max"]["latitude"],
          longitude: country["geographical"]["latLng_max"]["longitude"]),
      area: country["geographical"]["area"].toDouble(),
      region: country["geographical"]["region"],
      subregion: country["geographical"]["subregion"],
      region_code: country["geographical"]["region_code"],
      subregion_code: country["geographical"]["subregion_code"],
      postal_code: country["geographical"]["postal_code"],
      world_region: country["geographical"]["world_region"],
      landlocked: country["geographical"]["landlocked"],
      independent: country["geographical"]["independent"],
      continent: country["geographical"]["continent"]
          .map<XContinents>((element) => XContinents.values.singleWhere((continent) => element == continent.value))
          .toList(),
    );
  }

  void writeJsonFile(String path, Object object) {
    final File file = File(path);
    file.writeAsStringSync(json.encode(object));
    print("${object.runtimeType} exported at $path");
  }

  void printJson(Object object) {
    print(JsonEncoder.withIndent("  ").convert(object));
  }

  void unitialize() {
    print("XCountires is unitialized");
  }
}




// void initialize() {
//   print("XCountires is intialized");
// }

// void unitialize() {
//   print("XCountires is unitialized");
// }
